/*
 * (C) Copyright 2011-2013 JST - Japan Science and Technology Agency
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */

#include <linux/module.h>        // required by all modules
#include <linux/kernel.h>        // required by printk()
#include <linux/proc_fs.h>       // create_proc_entry
#include <asm/uaccess.h>         // copy_from_user
#include <linux/pid_namespace.h> // task_active_pid_ns
#include <linux/list.h>
#include <linux/cgroup.h>
#include <linux/ctype.h>

MODULE_LICENSE("GPL");
MODULE_AUTHOR("JST DEOS R&D Center");

#define PROCNAME "daware"

#define MAXBUF 64
static char _buf[MAXBUF];

#define MAX_ARGS 3

struct pid_ns_link {
    struct list_head pid_ns_link_list;
    pid_t pid;
    struct pid_namespace *ns;
    char *name;
};

struct args {
    int delete;
    int pid;
    char *name;
};

static DEFINE_SPINLOCK(lock);

struct list_head pid_ns_link_root;

#ifdef LATER_LINUX_2_6_34
static int
get_container_path(struct task_struct *tsk, char *buf, int len)
{
    struct cgroup_subsys_state *css;
    int retval;

    retval = -EINVAL;
    cgroup_lock();
    css = task_subsys_state(tsk, cpuset_subsys_id);
    retval = cgroup_path(css->cgroup, buf, len);
    cgroup_unlock();

    return retval;

}
#endif

static void
pid_ns_link_free(struct pid_ns_link *link)
{
    if (link) {
        if (link->name) kfree(link->name);
        kfree(link);
    }
}

static int
scan_args(char *buf, struct args *args) {
    int i;
    int argc;
    char *argv[MAX_ARGS] = { NULL };
    char *p;

    if (!args) return -EFAULT;
    memset(args, 0, sizeof(struct args));

    for (p = buf; *p; p++) {
        if (!isprint(*p)) *p = ' ';
    }

    i = 0;
    while (((p = strsep(&buf, " ")) != NULL) && (i < MAX_ARGS)) {
        if (*p) argv[i++] = p;
    }
    argc = i;

    if (argc <= 0) {
        return -EINVAL;
    }

    i = 0;
    if (!strcmp(argv[i], "-d")) {
        args->delete = 1;
        i++;
    }

    if ((argc - i) <= 0) {
        return -EINVAL;
    } else if ((argc - i) == 1) {
        args->pid = simple_strtoul(argv[i], NULL, 0);
    } else {
        args->name = argv[i++];
        args->pid = simple_strtoul(argv[i], NULL, 0);
    }

    if (!args->pid) {
        return -EINVAL;
    }

    return 0;
}

static int
daware_register(pid_t pid, const char *name)
{
    struct pid_namespace *ns;
    const struct task_struct *task;
    struct pid_ns_link *link;

    ns = current->nsproxy->pid_ns;

    list_for_each_entry(link, &pid_ns_link_root, pid_ns_link_list) {
        if (link->pid == pid && link->ns == ns) {
            return 0;
        }
    }

    task = pid_task(find_pid_ns(pid, ns), PIDTYPE_PID);
    if (!task || task->flags & PF_EXITING) {
        return 0;
    }

    link = kmalloc(sizeof(struct pid_ns_link), GFP_KERNEL);
    if (!link) {
        return -ENOMEM;
    }

    link->pid = pid;
    link->ns  = ns;
    link->name = name ? kstrdup(name, GFP_KERNEL) : NULL;

    list_add(&link->pid_ns_link_list, &pid_ns_link_root);

    return 0;
}

static int
daware_unregister(pid_t pid, const char *name)
{
    struct pid_namespace *ns;
    struct pid_ns_link *link;

    ns = current->nsproxy->pid_ns;

    list_for_each_entry(link, &pid_ns_link_root, pid_ns_link_list) {
        if (link->pid == pid && link->ns == ns
            && (!link->name || !name || !strcmp(link->name, name))) {

            list_del(&link->pid_ns_link_list);
            pid_ns_link_free(link);
            break;
        }
    }
    return 0;
}

static int
daware_write_proc(struct file *filp, const char *buf, unsigned long len, void *data)
{
    int ret;
    struct args args;

    if(len >= MAXBUF){
        return -ENOSPC;
    }

    spin_lock(&lock);

    if(copy_from_user(_buf, buf, len)) {
        spin_unlock(&lock);
        return -EFAULT;
    }
    _buf[len] = '\0';

    if (scan_args(_buf, &args)) {
        spin_unlock(&lock);
        return -EINVAL;
    }
    printk( KERN_INFO "DAWARE: %s %s %d\n",
            args.delete ? "delete" : "add",
            args.name ? args.name : "",
            args.pid);

    if (args.delete) {
        if ((ret = daware_unregister(args.pid, args.name))) {
            spin_unlock(&lock);
            return ret;
        }
    } else {
        if ((ret = daware_register(args.pid, args.name))) {
            spin_unlock(&lock);
            return ret;
        }
    }

    spin_unlock(&lock);

    return len;
}

static int
daware_read_proc(char *page, char **start, off_t offset, int count, int *eof, void *data)
{
    int n = 0;
    struct task_struct *task;
    struct pid_namespace *cur_ns, *act_ns;
    pid_t cur_nr, act_nr;
    struct pid_ns_link *link;
    struct pid_ns_link *save_link;

    spin_lock(&lock);

    n += sprintf(&page[n], "%-8s  %6s  %6s\n", "NAME", "PID", "VPID");
    list_for_each_entry_safe(
        link, save_link, &pid_ns_link_root, pid_ns_link_list) {

        task = pid_task(find_pid_ns(link->pid, link->ns), PIDTYPE_PID);
        if (!task || task->flags & PF_EXITING) {
            list_del(&link->pid_ns_link_list);
            pid_ns_link_free(link);
            continue;
        }

        cur_ns = current->nsproxy->pid_ns;
        cur_nr = cur_ns ? task_pid_nr_ns(task, cur_ns) : 0;
        if (!cur_nr) continue;

        act_ns = task_active_pid_ns(task);
        act_nr = act_ns ? task_pid_nr_ns(task, act_ns) : 0;

        n += sprintf(&page[n], "%-8s  %6d  %6d\n",
                     link->name ? link->name : "", cur_nr, act_nr);
    }

    *eof = 1;

    spin_unlock(&lock);

    return n;
}

// Start/Init function
static int daware_init(void) {
    struct proc_dir_entry *entry;

    printk( KERN_INFO "DAWARE: init\n");

    entry = create_proc_entry(PROCNAME, 0666, NULL);
    if (entry) {
        entry->write_proc = daware_write_proc;
        entry->read_proc  = daware_read_proc;
    } else {
        return -EBUSY;
    }

    INIT_LIST_HEAD(&pid_ns_link_root);

    return 0;
}

// End/Cleanup function
static void daware_exit(void) {
    struct pid_ns_link *link;
    struct pid_ns_link *save_link;

    printk( KERN_INFO "DAWARE: exit\n");

    remove_proc_entry(PROCNAME, NULL);

    spin_lock(&lock);
    list_for_each_entry_safe(
        link, save_link, &pid_ns_link_root, pid_ns_link_list) {

        list_del(&link->pid_ns_link_list);
        pid_ns_link_free(link);
    }
    spin_unlock(&lock);
}

module_init(daware_init);
module_exit(daware_exit);
