#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdarg.h>
#include <time.h>
#include <sys/time.h>

#define SYSLOG_NAMES //for facilitynames and prioritynames in syslog.h
#include <syslog.h>

#include "daware.h"

extern const char*
get_container_name();

#define DEFAULT_FACILITY LOG_LOCAL5
#define DEFAULT_FORMAT   "%msg%"
//#define DEFAULT_FORMAT   "container=%cont%;%msg%"


#define DEBUG_LOG(fmt, ...)
//#define DEBUG_LOG(fmt, ...) printf(fmt, ##__VA_ARGS__)

#define LOGFMT_KEY_FACI  0
#define LOGFMT_KEY_LVL   1
#define LOGFMT_KEY_TIME  2
#define LOGFMT_KEY_HOST  3
#define LOGFMT_KEY_PROG  4
#define LOGFMT_KEY_PID   5
#define LOGFMT_KEY_MSG   6
#define LOGFMT_KEY_FILE  7
#define LOGFMT_KEY_LINE  8
#define LOGFMT_KEY_FUNC  9
#define LOGFMT_KEY_CONT  10

typedef struct {
    const char *str;
    int id;
} logfmt_key_t;

logfmt_key_t logfmt_keys[] = {
    {"%faci%", LOGFMT_KEY_FACI},
    {"%lvl%",  LOGFMT_KEY_LVL},
    {"%time%", LOGFMT_KEY_TIME},
    {"%host%", LOGFMT_KEY_HOST},
    {"%prog%", LOGFMT_KEY_PROG},
    {"%pid%",  LOGFMT_KEY_PID},
    {"%msg%",  LOGFMT_KEY_MSG},
    {"%file%", LOGFMT_KEY_FILE},
    {"%line%", LOGFMT_KEY_LINE},
    {"%func%", LOGFMT_KEY_FUNC},
    {"%cont%", LOGFMT_KEY_CONT},
    {NULL,     -1}
};
#define LOGFMT_KEY_MAX (sizeof(logfmt_keys)/sizeof(logfmt_keys[0]) - 1)

static int default_facility = DEFAULT_FACILITY;
static void
set_default_facility(int facility)
{
    if ((facility & LOG_FACMASK) == 0) {
        facility = DEFAULT_FACILITY;
    }
    default_facility = facility & LOG_FACMASK;
}

static int
get_default_facility()
{
    return default_facility;
}

static char logfmt_buf[256] = DEFAULT_FORMAT;
static void
set_logfmt(const char *logfmt)
{
    memset(logfmt_buf, 0, sizeof(logfmt_buf));
    strncpy(logfmt_buf, logfmt, sizeof(logfmt_buf) - 1);
}

static const char*
get_logfmt()
{
    return logfmt_buf;
}

static const char*
get_progname_str()
{
    static int is_init = 0;
    static char progname[256] = { 0 };

    if (!is_init) {
        is_init = 1;
        FILE *fp = fopen("/proc/self/status","r");
        if (fp) {
            if (fscanf(fp, "Name:%255s", progname) != 1) {
                *progname = '\0';
            }
            fclose(fp);
        }
    }

    return progname;
}

static char*
get_time_str(char* buf, int size)
{
    struct timeval tv;
    if (gettimeofday(&tv, NULL) == -1) {
        perror("gettimeofday");
        return "";
    }
    struct tm *t = localtime(&tv.tv_sec);
    if (t == NULL) {
        fprintf(stderr, "localtime: return null.");
        return "";
    }
    strftime(buf, size, "%b %d %H:%M:%S", t);
    snprintf(&buf[strlen(buf)], size - strlen(buf), ".%ld", tv.tv_usec / 1000);

    return buf;
}

static const char*
get_hostname_str()
{
    static int is_init = 0;
    static char buf[256];
    if (!is_init) {
        is_init = 1;
        if (gethostname(buf, sizeof(buf)) == -1) {
            perror("gethostname");
            buf[0] ='\0';
        }
    }
    return buf;
}

static const char*
get_logfmt_key(int key)
{
    int i;
    for (i = 0; logfmt_keys[i].id != -1; i++) {
        if (key == logfmt_keys[i].id) {
            return logfmt_keys[i].str;
        }
    }
    return "";
}

static const char*
get_facility_str(int priority)
{
    if (LOG_FAC(priority) == 0) {
        priority = get_default_facility();
    }

    int i;
    for (i = 0; facilitynames[i].c_val != -1; i++) {
        if (LOG_FAC(priority) == LOG_FAC(facilitynames[i].c_val)) {
            return facilitynames[i].c_name;
        }
    }
    return "";
}

static const char*
get_level_str(int priority)
{
    int i;
    for (i = 0; prioritynames[i].c_val != -1; i++) {
        if (LOG_PRI(priority) == LOG_PRI(prioritynames[i].c_val)) {
            return prioritynames[i].c_name;
        }
    }
    return "";
}

static char*
int_to_str(char* buf, int size, int i)
{
    snprintf(buf, size, "%d", i);
    return buf;
}

static int is_open = 0;
void
daware_log_open(const char *ident, int option, int facility)
{
    is_open = 1;
    openlog(ident, option, facility);
    set_default_facility(facility);
}

void
daware_log_close()
{
    closelog();
}

void
daware_log_set_format(const char *logfmt)
{
    set_logfmt(logfmt);
}

void
daware_log(int priority,
             const char *file, int line, const char *func,
             const char *fmt, ...)
{
    if (is_open == 0) {
        is_open = 1;
        openlog(get_progname_str(), LOG_PID, DEFAULT_FACILITY);
    }

    if ((priority & LOG_FACMASK) == 0) {
        priority |= DEFAULT_FACILITY;
    }

    char msg[1024];
    va_list ap;
    va_start(ap, fmt);
    vsnprintf(msg, sizeof(msg), fmt, ap);
    va_end(ap);

    char linebuf[16], pidbuf[16], timebuf[64];
    const char *param[LOGFMT_KEY_MAX];
    param[LOGFMT_KEY_FACI] = get_facility_str(priority);
    param[LOGFMT_KEY_LVL]  = get_level_str(priority);
    param[LOGFMT_KEY_TIME] = get_time_str(timebuf, sizeof(timebuf));
    param[LOGFMT_KEY_HOST] = get_hostname_str();
    param[LOGFMT_KEY_PROG] = get_progname_str();
    param[LOGFMT_KEY_PID]  = int_to_str(pidbuf, sizeof(pidbuf), getpid());
    param[LOGFMT_KEY_MSG]  = msg;
    param[LOGFMT_KEY_FILE] = file ? file : "";
    param[LOGFMT_KEY_LINE] = int_to_str(linebuf, sizeof(linebuf), line);
    param[LOGFMT_KEY_FUNC] = func ? func : "";
    param[LOGFMT_KEY_CONT] = get_container_name();

    char log[1024];
    memset(log, 0, sizeof(log));

    char *plog = log;
    char *plog_limit = log + sizeof(log) - 1;

    const char *pfmt = get_logfmt();

    while(strlen(pfmt) && 0 < plog_limit - plog) {
        int ident;
        for (ident = 0; ident < LOGFMT_KEY_MAX; ident++) {
            const char *key = get_logfmt_key(ident);
            if (strncmp(pfmt, key, strlen(key)) == 0) {
                int size = strlen(param[ident]);
                if (plog_limit - plog < size) size = plog_limit - plog;
                if (0 < size) {
                    strncpy(plog, param[ident], size);
                    plog += size;
                }
                pfmt += strlen(key);
                continue;
            }
        }
        *plog++ = *pfmt++;
    }

    DEBUG_LOG("log:[%s]\n", log);

    syslog(priority, "%s\n", log);
}
